module dfc.examples.bitmapview.bitview;

import win32.api;
import dfc.bitmap;
import dfc.core;
import dfc.button;
import dfc.tree;
import dfc.edit;
import winex.winmm;
import win32.commdlg;

import dfc.examples.bitmapview.layout;
import dfc.examples.bitmapview.abstapp;
import dfc.examples.bitmapview.basecontrols;
import dfc.examples.bitmapview.mycontrols;

public class BitViewApp : BaseApp {

	alias UserControl!(BitViewApp) UserCtrl;

	DCachedBitmap bitmap;

	class BitmapViewWnd : UserCtrl.UserDefinedControl, UserScrollable, HasNotifyOnSize {
		int zoom = 1;

		this( BitViewApp app0 ) {
			super( app0 );
			addPaintHandler( WM_PAINT, &onPaint );
		}

		LRESULT onPaint( HDC dc ) {
			RECT rect;
			int img_w, img_h, scrn_w, scrn_h, v_pos, h_pos, ox=0, oy=0;
			getClientRect( rect );

			DCachedBitmap bm = app.bitmap;
			if ( !bm ) {
				FillRect( dc, &rect, app.getMidlightTwoBrush() );
				return cast(LRESULT)0;
			}

			bm.getZoomParam( zoom, img_w, img_h );
			scrn_w = rect.right - rect.left;
			scrn_h = rect.bottom - rect.top;
			h_pos = app.hscroller.getPosition();
			v_pos = app.vscroller.getPosition();

			if ( scrn_w > img_w ) {
				RECT fill;
				ox = ((scrn_w-img_w)>>1);
				fill.left = rect.left;
				fill.right = rect.left + ox;
				fill.top = rect.top;
				fill.bottom = rect.bottom;
				FillRect( dc, &fill, app.getMidlightOneBrush() );
				fill.left = rect.left + (ox+img_w);
				fill.right = rect.right;
				FillRect( dc, &fill, app.getMidlightOneBrush() );
				rect.left = ox;
				rect.right = ox+img_w;
			} else {
				ox = -h_pos;
			}
			if ( scrn_h > img_h ) {
				RECT fill;
				oy = ((scrn_h-img_h)>>1);
				fill.left = rect.left;
				fill.right = rect.right;
				fill.top = rect.top;
				fill.bottom = rect.top + oy;
				FillRect( dc, &fill, app.getMidlightOneBrush() );
				fill.top = rect.top + (oy+img_h);
				fill.bottom = rect.bottom;
				FillRect( dc, &fill, app.getMidlightOneBrush() );
				rect.top = oy;
				rect.bottom = oy+img_h;
			} else {
				oy = -v_pos;
			}

			bm.renderScaledOnto( dc, ox, oy, img_w, img_h );

			return cast(LRESULT)0;
		}

		void newBitmap( int newzoom ) {
			if ( newzoom != 0 ) {
				zoom = newzoom;
			}
			DCachedBitmap bm = app.bitmap;
			if ( bm ) {
				RECT rect;
				int img_w, img_h, scrn_w, scrn_h, v_pos, h_pos;
				getClientRect( rect );
				bm.getZoomParam( zoom, img_w, img_h );
				scrn_w = rect.right - rect.left;
				scrn_h = rect.bottom - rect.top;
				h_pos = app.hscroller.getPosition();
				v_pos = app.vscroller.getPosition();
				app.hscroller.setParameters( 0, img_w, bm.getWidth()>>4, h_pos, scrn_w  );
				app.vscroller.setParameters( 0, img_h, bm.getHeight()>>4, v_pos, scrn_h );
			} else {
				app.hscroller.setParameters( 0, 10, 1, 0, 10 );
				app.vscroller.setParameters( 0, 10, 1, 0, 10 );
			}

			invalidate();
		}

 /* interface UserScrollable { */
		void dragStarted( int pos, UserScrollMode mode, UserScroller scroller )	{
		}
		void dragUpdate(  int pos, UserScroller scroller ) {
			invalidate();
		}
		void dragEnded(   int pos, UserScrollMode mode, UserScroller scroller ) {
			invalidate();
		}
 /* } interface UserScrollable */

	/* interface HasNotifyOnSize { */
		void newSizeSet( int width, int height ) {
			int v_pos = app.vscroller.getPosition();
			int h_pos = app.hscroller.getPosition();
			if ( app.bitmap ) {
				int img_w, img_h;
				app.bitmap.getZoomParam( zoom, img_w, img_h );
				app.hscroller.setParameters( 0, img_w, 8*zoom, h_pos, width  );
				app.vscroller.setParameters( 0, img_h, 8*zoom, v_pos, height );
			} else {
				app.hscroller.setParameters( 0, width,  8, h_pos, width  );
				app.vscroller.setParameters( 0, height, 8, v_pos, height );
			}
			invalidate();
		}
	/* } HasNotifyOnSize */
	}

	UserCtrl.UserScrollbar hscroller, vscroller;
	BitmapViewWnd bitmapview;

	this() {
		super();
		create(	WS_THICKFRAME | WS_MAXIMIZEBOX | WS_MINIMIZEBOX | WS_SYSMENU|WS_CLIPCHILDREN,
					0,
					"Bitmap Viewer",
					500, 400 );
		BorderLayout borderlayout = new BorderLayout( true );
		ScrollBorderLayout scrolllayout = new ScrollBorderLayout();
		RowLayout    buttons = new RowLayout();
		/* connect the root layout to this window */
		LayoutManager.connect( this, borderlayout );

//		addPaintHandler( WM_PAINT, &onPaint ); // only needed if you wish to paint on this the parent window

		bitmapview = new BitmapViewWnd( this );
		hscroller  = new UserCtrl.UserScrollbar( this );
		vscroller  = new UserCtrl.UserScrollbar( this, true );
		/* tell the scroll bars to notify the bitmap view window not this window */
		hscroller.setNotify( bitmapview );
		vscroller.setNotify( bitmapview );
		/* set up the scroll layout*/
		scrolllayout.add( hscroller,"horiz", 100, 12 );
		scrolllayout.add( vscroller, "vert", 12, 100 );
		scrolllayout.add( new UserCtrl.UserFill( this ), "edge", 12, 12 );
		scrolllayout.add( bitmapview, "center", 100, 100 );

		/* setup the root layout */
		borderlayout.add( buttons, "north", 100, 25 );
		borderlayout.add( scrolllayout, "center", 100, 25 );

		buttons.add( new UserCtrl.UserButton( this, "Load", &onLoad ), "", 100, 24 );
		buttons.add( new UserCtrl.UserButton( this, "25%", &on25pc ), "", 100, 24 );
		buttons.add( new UserCtrl.UserButton( this, "50%", &on50pc ), "", 100, 24 );
		buttons.add( new UserCtrl.UserButton( this, "100%", &on100pc  ), "", 100, 24 );
		buttons.add( new UserCtrl.UserButton( this, "200%", &on200pc  ), "", 100, 24 );
		buttons.add( new UserCtrl.UserButton( this, "400%", &on400pc  ), "", 100, 24 );

		centerWindow();

		hscroller.setParameters( 0, 10, 1, 0, 10 );
		vscroller.setParameters( 0, 10, 1, 0, 10 );
	}

	void preCreateWindow( inout DWORD dwStyle, inout DWORD dwExStyle )
	{
		dwStyle |= WS_CLIPSIBLINGS|WS_CLIPCHILDREN;
	}
/* // not needed for this application all of this window is cover by child windows
	LRESULT onPaint( HDC dc ) {
		RECT rect;
		getClientRect( rect );
		FillRect( dc, &rect, getMidlightOneBrush() );
		return (LRESULT)0;
	}
*/

	private void onLoad( UserCtrl.UserButton button ) {
		char[] filename;
/*  // if using this from a child window or dialog use.
		if ( FileLoadDialog.openFile( app.getHwnd(), filename ) ) {
*/
		if ( FileLoadDialog.openFile( getHwnd(), filename ) ) {
			DCachedBitmap bm = new DCachedBitmap( filename );
			if ( bm.getWidth() < 0 ) {
				MessageBox( hwnd, "unable to load requested file:'"~filename~"'", "Load Error", MB_OK | MB_TASKMODAL );
			} else {
				bitmap = bm;
				bitmapview.newBitmap( 0 );
			}
		} else {
			printf( "LOAD nothing to load\n" );
		}
//		return cast(LRESULT)0; // if used from within a mesage handler use this line
	}

	private void onSave( UserCtrl.UserButton button ) {
		char[] filename;
/*  // if using this from a child window or dialog use.
		if ( FileSaveDialog.openFile( app.getHwnd(), filename ) ) {
*/
		if ( FileSaveDialog.openFile( getHwnd(), filename ) ) {
			printf( "Save to %.*s\n", filename );
		} else {
			printf( "Save did nothing\n" );
		}
//		return cast(LRESULT)0; // if used from within a mesage handler use this line
	}


	private void on25pc( UserCtrl.UserButton button ) {
		bitmapview.newBitmap( -4 );
	}
	private void on50pc( UserCtrl.UserButton button ) {
		bitmapview.newBitmap( -2 );
	}
	private void on100pc( UserCtrl.UserButton button ) {
		bitmapview.newBitmap( 1 );
	}
	private void on200pc( UserCtrl.UserButton button ) {
		bitmapview.newBitmap( 2 );
	}
	private void on400pc( UserCtrl.UserButton button ) {
		bitmapview.newBitmap( 4 );
	}
}
