module dfc.examples.bitmapview.layout;

import win32.api;
import dfc.bitmap;
import dfc.core;
import dfc.button;
import dfc.tree;
import dfc.edit;
import winex.winmm;
import win32.commdlg;

import dfc.examples.bitmapview.baselayout;


//import mapctrls;
//version = LAYOUT_DEBUG;
version = LAYOUT_RELEASE;

version (LAYOUT_DEBUG) {
	import std.c.stdio;
}

public class LayoutItem {
	wchar[] constraint;
	this( wchar[] cst ) { constraint = cst; }
	void reshape( int x, int y, int w, int h ) { }
	void getSize( out int w, out int h ) { }
	void enable( bit enabled ) { }
}

public class WndLayout : LayoutItem {
	DWnd parent;
	this( DWnd wnd, wchar[] cst ) { super( cst ); parent = wnd; }
	void reshape( int x, int y, int w, int h ) {
		parent.moveWindow( x, y, w, h, true );
		HasNotifyOnSize sizer;
		sizer = cast(HasNotifyOnSize)parent;
		if ( sizer ) {
			sizer.newSizeSet( w, h );
		}
	}
	void getSize( out int w, out int h ) {
		HasPreferedSize sizer;
		sizer = cast(HasPreferedSize)parent;
		if ( sizer ) {
			sizer.getPreferedSize( w, h );
		} else {
			RECT r;
			parent.getWindowBounds( r );
			w = r.right;
			h = r.bottom;
		}
	}
	void enable( bit enabled ) { parent.showWindow( enabled ); }
}

public class NestedLayout : LayoutItem {
	LayoutManager parent;
	int width, height;
	this( LayoutManager mgr, wchar[] cst ) { super( cst ); parent = mgr; }
	void reshape( int x, int y, int w, int h ) {
		parent.resize( x, y, w, h );
		width  = w;
		height = h;
	}
	void getSize( out int w, out int h ) {
		w = width;
		h = height;
	}
}

public class LayoutManager {

	public this() { }

	public final void add( DWnd child, wchar[] constraints ) {
		addi( new WndLayout( child, constraints ) );
	}

	public final void add( LayoutManager child, wchar[] constraints ) {
		addi( new NestedLayout( child, constraints ) );
	}

	public final void add( DWnd child, wchar[] constraints, int w, int h ) {
		WndLayout wl;
		addi( wl = new WndLayout( child, constraints ) );
		wl.reshape( 0,0,w, h );
	}

	public final void add( LayoutManager child, wchar[] constraints, int w, int h ) {
		NestedLayout nl;
		addi( nl = new NestedLayout( child, constraints ) );
		nl.reshape(0,0, w, h );
	}

	protected void addi( LayoutItem item ) {
		throw new Exception( "LayoutManager::addi(LayoutItem) must be defined" );
	}

	public void resize( int x, int y, int w, int h ) {
		throw new Exception( "LayoutManager::resize(int,int,int,int) must be defined" );
	}

	static void connect( DWnd wnd, LayoutManager mgr ) {
		WndLayoutController.connect( wnd, mgr );
	}
}

public class WndLayoutController {
	protected DWnd parent;
	protected LayoutManager manager;
	this( DWnd p, LayoutManager mgr ) {
		parent  = p;
		manager = mgr;
		p.addHandler( WM_SIZE, new ResizeMessageSlot( &onResize ) );
	}

	private final LRESULT onResize( int width, int height ) {
//		RECT r;
//		parent.getWindowPosition( r );
//		manager.resize( r.left, r.top, r.right - r.left, r.bottom - r.top );
		manager.resize( 0, 0, width, height );
		return cast(LRESULT)0;
	}

	static void connect( DWnd wnd, LayoutManager mgr ) {
		new WndLayoutController( wnd, mgr );
	}
}

public class ListLayout : LayoutManager {
	LayoutItem[] itemlist;
	protected void addi( LayoutItem item ) {
		itemlist ~= item;
	}
}

public class RowLayout : ListLayout {
	public void resize( int x, int y, int w, int h ) {
		int idx, len;
		len = itemlist.length;
		idx = 0;
		while( idx < len ) {
			int iw;
			iw = w / (len - idx);
			itemlist[idx].reshape( x, y, iw, h );
			x += iw;
			w -= iw;
			idx++;
		}
	}
}

public class ColumnLayout : ListLayout {
	public void resize( int x, int y, int w, int h ) {
		int idx, len;
		len = itemlist.length;
		idx = 0;
		while( idx < len ) {
			int ih;
			ih = h / (len - idx);
			itemlist[idx].reshape( x, y, w, ih );
			y += ih;
			h -= ih;
			idx++;
		}
	}
}

public class BorderLayout : LayoutManager {
	LayoutItem center;
	LayoutItem north;
	LayoutItem south;
	LayoutItem east;
	LayoutItem west;
	bit resizeEWfirst;
	this( bit ewf ) { resizeEWfirst = ewf; }
	this() { this( false ); }

	protected void addi( LayoutItem item ) {
		switch( item.constraint ) {
		case "center": center = item; break;
		case  "north":  north = item; break;
		case  "south":  south = item; break;
		case   "east":   east = item; break;
		case   "west":   west = item; break;
		default:
			throw new Exception( "invalid item constraints" );
		}
		version (LAYOUT_DEBUG) {
			printf( "Adding layout item %.*s\n", item.constraint );
		}
	}

	public void resize( int x, int y, int w, int h ) {
		int iw, ih;

		void resizeNS( inout int x, inout int y, inout int w, inout int h ) {
			if ( north ) {
				north.getSize( iw, ih );
				north.reshape( x, y, w, ih );
				version (LAYOUT_DEBUG) {
					printf( "north->(%d, %d, %d, %d)...\n", x, y, w, ih );
				}
				y += ih; h -= ih;
			}
			if ( south ) {
				south.getSize( iw, ih );
				south.reshape( x, y + h - ih, w, ih );
				version (LAYOUT_DEBUG) {
					printf( "south->(%d, %d, %d, %d)...\n", x, y + h - ih, w, ih );
				}
				h -= ih;
			}
		}

		void resizeEW( inout int x, inout int y, inout int w, inout int h ) {
			if ( west ) {
				west.getSize( iw, ih );
				west.reshape( x, y, iw, h );
				version (LAYOUT_DEBUG) {
					printf( "west->(%d, %d, %d, %d)...\n", x, y, iw, h );
				}
				x += iw; w -= iw;
			}
			if ( east ) {
				east.getSize( iw, ih );
				east.reshape( x + w -iw, y, iw, h );
				version (LAYOUT_DEBUG) {
					printf( "east->(%d, %d, %d, %d)...\n", x +w-iw, y, iw, h );
				}
				w -= iw;
			}
		}

		if ( resizeEWfirst ) {
			resizeEW( x, y, w, h );
			resizeNS( x, y, w, h );
		} else {
			resizeNS( x, y, w, h );
			resizeEW( x, y, w, h );
		}
		if ( center ) {
			center.reshape( x, y, w, h );
			version (LAYOUT_DEBUG) {
				printf( "center->(%d, %d, %d, %d)...\n", x, y, w, h );
			}
		}
		version (LAYOUT_DEBUG) {
			printf( "resized...\n" );
		}
	}
}

public class TabLayout : ListLayout {
	LayoutItem parent;
	int tab_height;
	int indent;
	int idx;
	bit tab_at_top;
	this( int h, int idt, bit pos ) {
		tab_height = h;
		indent = idt;
		tab_at_top = pos;
		idx = 0;
	}
	this( int h, int idt ) { this ( h, idt, true ); }
	this( bit pos ) { this( 28, 3, pos ); }
	this() { this( 28, 3, true ); }

	public void setIndex( int i ) { idx = i; }

	public void setLayoutParams( int h, int idt, bit at_top ) {
		tab_height = h;
		indent   = idt;
		tab_at_top = at_top;
	}

	protected void addi( LayoutItem item ) {
		switch( item.constraint ) {
		case "parent": parent = item; break;
		default:
			super.addi( item );
		}
	}

	public void resize( int x, int y, int w, int h ) {
		if ( parent) { parent.reshape( x, y, w, h ); }
		if ( idx < itemlist.length ) {
			for ( int i = 0; i < itemlist.length; i++ ) {
				// assume itemlist[n] is a child of parent!
				// if not then this is the wrong type of layout manager
				if ( tab_at_top ) {
					itemlist[i].reshape( indent, tab_height, w - (indent*3), (h-(tab_height+(indent<<1))) );
				} else {
					itemlist[i].reshape( indent, indent, w - (indent<<1), (h-(tab_height+indent)) );
				}
				itemlist[i].enable( i == idx );
			}
		}
	}
}

public class ScrollBorderLayout : LayoutManager {
	LayoutItem horiz;
	LayoutItem vert;
	LayoutItem center;
	LayoutItem edge;
	this() { }

	protected void addi( LayoutItem item ) {
		switch( item.constraint ) {
		case "center": center = item; break;
		case  "horiz":  horiz = item; break;
		case   "vert":   vert = item; break;
		case   "edge":   edge = item; break;
		default:
			throw new Exception( "invalid item constraints" );
		}
	}

	public void resize( int x, int y, int w, int h ) {
		int ignore, cw, ch;
		cw = ch = 0;
		if ( edge ) {
			edge.getSize( cw, ch );
		}
		if ( horiz ) {
			horiz.getSize( ignore, ch );
		}
		if ( vert ) {
			vert.getSize( cw, ignore );
			vert.reshape( x + w - cw, y, cw, h - ch );
		}
		if ( horiz ) {
			horiz.reshape( x, y + h - ch, w-cw, ch );
		}
		if ( edge ) {
			edge.reshape( x + w - cw, y + h - ch, cw, ch );
		}

		if ( center ) {
			center.reshape( x, y, w - cw, h - ch );
		}
	}
}

