module dfc.examples.bitmapview.mycontrols;

import std.c.stdio;

import win32.api;
import dfc.bitmap;
import dfc.core;
import dfc.examples.bitmapview.basecontrols;

import dfc.examples.bitmapview.baselayout;


interface UserScroller {
	int getPosition();
	int getRange();
	int getMin();
	int getMax();
	int getPage();
	void getMaxMin( out int min, out int max );
	void getPosRange( out int pos, out int range );

	void setPosition( int pos );
	void setRange(    int range );
	void setMin(      int min );
	void setMax(      int max );
	void setPage(     int page );

	void setMaxMin( int min, int max );
	void setMaxMinPage( int min, int max, int page );
	void setPosRange( int pos, int range );
	void setParameters( int min, int max, int page, int pos, int range );
}

enum UserScrollMode {
	DRAG, PAGE
}

interface UserScrollable {
	void dragStarted( int pos, UserScrollMode mode, UserScroller scroller );
	void dragUpdate(  int pos, UserScroller scroller );
	void dragEnded(   int pos, UserScrollMode mode, UserScroller scroller );
}

public template UserControl( Application : UserApp ) {
	alias Control!( Application ) AppControl;

	class UserDefinedControl : AppControl.AppWnd {

		this( Application parent, AppControl.AppWnd.CreateDelegate cf ) {
			super( parent, cf);
		}
		this( Application parent ) { super( parent ); }

		protected this( Application parent, char[] clazz ) {
			super( parent, clazz );
		}
		protected this( Application app0, DWnd parent ) {
			super( app0, parent );
		}
		protected this( Application app0, DWnd parent, char[] clazz ) {
			super( app0, parent, clazz );
		}
		protected this( Application app0, DWnd parent, char[] clazz, int id ) {
			super( app0, parent, clazz, id );
		}
		protected this( bit fake ) {
			super( fake );
		}
	}

	class UserScrollbar : AppControl.AppWnd, UserScroller, HasPreferedSize {
	public:
		int min, max, pos, range, page;
		bit vertical;
		RECT button;
		bit mouse_active;
		bit drag_active;
		bit focus;
		int page_dir, ox, oy, start_pos;
		UserScrollable scrollable;
		this( Application app0, DWnd parent, bit vert ) {
			super( app0, parent );
			addPaintHandler( WM_PAINT, &onPaint );
			addMouseDownHandler( WM_LBUTTONDOWN, &onMouseDown );
			addMouseUpHandler(   WM_LBUTTONUP,   &onMouseUp ); // release capture
			addMouseDownHandler( WM_RBUTTONDOWN, &onMouseDown );
			addMouseUpHandler(   WM_RBUTTONUP,   &onMouseUp ); // release capture
			addMouseHandler( WM_MOUSEMOVE, &onMouseMove );
			addFocusHandler( WM_SETFOCUS, &onFocusGained );
			addFocusHandler( WM_KILLFOCUS, &onFocusLost );
			min = 0; max = 100; pos = 0; range = 20; page = 10;
			vertical = vert;
			mouse_active = false;
			drag_active = false;
			scrollable = cast(UserScrollable)app0;
		}
		this( Application parent ) { this( parent, parent, false ); }
		this( Application parent, bit vert ) { this( parent, parent, vert ); }
		this( Application app0, DWnd parent ) { this( app0, parent, false ); }

		void setNotify( UserScrollable scrollable0 ) {
			scrollable = scrollable0;
		}

	/* interface HasPreferedSize { */
		void getPreferedSize( out int width, out int height ) {
			RECT r;
			getWindowBounds( r );
			if ( vertical ) {
				width = 12;
				height = r.bottom;
			} else {
				width = r.right;
				height = 12;
			}
//			printf("gps(%d,%d)\n", width, height );
		}
	/* } interface HasPreferedSize */

	/* interface UserScroller { */
		int getPosition() { return pos;   }
		int getRange()    { return range; }
		int getMin()      { return min;   }
		int getMax()      { return max;   }
		int getPage()     { return page;  }
		void getMaxMin( out int min0, out int max0 )	{
			min0 = min; max0 = max;
		}
		void getPosRange( out int pos0, out int range0 ) {
			pos0 = pos; range0 = range;
		}
		void setPosition( int pos0 )   { pos   = pos0;   validateParams(); }
		void setRange(    int range0 ) {	range = range0; validateParams(); }
		void setMin(      int min0 )	 { min   = min0;   validateParams(); }
		void setMax(      int max0 )	 { max   = max0;   validateParams(); }
		void setPage(     int page0 )  { page  = page0;  validateParams(); }
		void setMaxMin( int min0, int max0 ) {
			min = min0; max = max0;
			validateParams();
		}
		void setMaxMinPage( int min0, int max0, int page0 ) {
			min = min0; max = max0; page = page0;
			validateParams();
		}
		void setPosRange( int pos0, int range0 ) {
			if ( (pos0+range0) > max ) { pos0 = max - range0; }
			pos = pos0>min ? pos0:min;
			range = range0;
			validateParams();
		}
		void setParameters( int min0, int max0, int page0, int pos0, int range0 ) {
			min = min0; max = max0;
			pos = pos0;
			range = range0;
			page = page0;
			validateParams();
		}

		private void validateParams() {
			if ( (pos+range) > max ) { pos = max - range; }
			if ( pos < min ) pos = min;
			invalidate();
		}

	/* } interface UserScroller */

		private int side( inout RECT r, int x, int y ) {
			if ( vertical ) {
				if ( y < r.bottom ) {
					return y < r.top ? -1 : 0;
				}
				return 1;
			} else {
				if ( x < r.right ) {
					return x < r.left ? -1 : 0;
				}
				return 1;
			}
			return 0;
		}

		void drawLoweredBox( HDC dc, int x, int y, int w, int h ) {
			RECT r;
			HPEN oldPen = cast(HPEN)SelectObject( dc, app.getLolightOnePen() );
//			printf( "Box(%d,%d x %d,%d)\n",x,y,w,h );
			r.left = x+1; r.right  = x+w-1;
			r.top  = y+1; r.bottom = y+h-1;
			FillRect( dc, &r, app.getMidlightOneBrush() );
			MoveToEx( dc, x, y + h-1, null );
			LineTo(   dc, x, y );
			LineTo(   dc, x + w-1, y );
			SelectObject( dc, app.getMidlightOnePen() );
			LineTo(   dc, x + w-1, y+1 );
			SelectObject( dc, app.getHilightOnePen() );
			LineTo(   dc, x + w-1, y + h-1 );
			LineTo(   dc, x, y + h-1 );
			SelectObject( dc, app.getMidlightOnePen() );
			LineTo(   dc, x-1, y + h-1 );
			SelectObject( dc, oldPen );
		}

//		inherited "owner";
//		Application app;
		LRESULT onPaint( HDC dc ) {
			RECT rect;
//			printf( "UserScrollbar::onPaint\n" );
			getClientRect( rect );
			drawTricolBox( dc, rect,
								app.getLolightOnePen(),
								app.getMidlightOnePen(),
								app.getHilightOnePen(),
								app.getMidlightOneBrush() );
//			drawLoweredBox( dc, rect.left, rect.top, rect.right-rect.left, rect.bottom-rect.top );
			setDeflatedRect( button, rect, 1, 1 );
			if ( (pos <= min) && ((pos+range)>=max) ) {
				drawTricolBox( dc, button,
								app.getHilightOnePen(),
								app.getMidlightOnePen(),
								app.getLolightOnePen(),
								app.getMidlightOneBrush() );
/*				drawTricolBox( dc, button,
								app.getHilightThreePen(),
								app.getMidlightThreePen(),
								app.getLolightThreePen(),
								app.getMidlightThreeBrush() ); */
				return cast(LRESULT)0;
			}

			int fr, vw, sp;
			fr = max - min;

			if ( vertical ) {
				vw = (button.bottom - button.top);
				sp = button.top;
				button.top    = sp + (vw*pos)/fr;
				button.bottom = sp + (vw*(pos+range))/fr;
			} else {
				vw = (button.right - button.left);
				sp = button.left;
				button.left  = sp + (vw*pos)/fr;
				button.right = sp + (vw*(pos+range))/fr;
			}

			if ( focus ) {
				drawTricolBox( dc, button,
								app.getHilightThreePen(),
								app.getMidlightThreePen(),
								app.getLolightThreePen(),
								app.getMidlightThreeBrush() );
      	} else {
				drawTricolBox( dc, button,
								app.getHilightTwoPen(),
								app.getMidlightTwoPen(),
								app.getLolightTwoPen(),
								app.getMidlightTwoBrush() );
			}
			return cast(LRESULT)0;
		}

		LRESULT onMouseMove( int x, int y, int keys ) {
			if ( drag_active ) {
				RECT rect;
				int fr, vw, delta, op;
				op = pos;
				getClientRect( rect );
				deflateRect( rect, 1, 1 );
				fr = max - min;
				if ( vertical ) {
					vw = (rect.bottom - rect.top);
					delta = y-oy;
				} else {
					vw = (rect.right - rect.left);
					delta = x-ox;
				}
				pos = start_pos + (delta*fr)/vw;
				if ( pos < min ) { pos = min; }
				if ( pos > (max-range) ) { pos = max - range; }
				if ( pos != op ) {
					if ( scrollable ) {
						scrollable.dragUpdate(  pos, this );
					}
					invalidate();
				}
			}
			return cast(LRESULT)0;
		}

		LRESULT onMouseDown( int x, int y, int keys ) {
			mouse_active = true;
//			printf( "UserScrollbar::onMouseDown [enter %p]\n", &button );
			if ( (page_dir = side( button, x, y )) == 0 ) {
				drag_active = true;
				ox = x; oy = y; start_pos = pos;
			}
			if ( scrollable ) {
				scrollable.dragStarted(
									pos,
									drag_active ? UserScrollMode.DRAG : UserScrollMode.PAGE,
									this );
			}
			SetFocus( hwnd );
//			printf( "UserScrollbar::onMouseDown [exit]\n" );
			return cast(LRESULT)0;
		}

		LRESULT onMouseUp( int x, int y, int keys ) {
			if ( !drag_active ) {
				switch( page_dir ) {
				case -1:
					pos -= page;
					if ( pos < min ) { pos = min; }
					break;
				case 1:
					pos += page;
					if ( pos > (max-range) ) { pos = max-range; }
					break;
				case 0:
				default:
					break;
				}
				invalidate();
			}
			if ( scrollable ) {
				scrollable.dragEnded(
									pos,
									drag_active ? UserScrollMode.DRAG : UserScrollMode.PAGE,
									this );
			}
			mouse_active = false;
			drag_active = false;
			return cast(LRESULT)0;
		}

		LRESULT onFocusLost( HWND oldWnd ) {
			focus = false;
			invalidate();
			return cast(LRESULT)0;
		}

		LRESULT onFocusGained( HWND oldWnd ) {
			focus = true;
			invalidate();
			return cast(LRESULT)0;
		}
	}
	class UserButton : AppControl.AppWnd, HasPreferedSize {
		alias void delegate( UserButton ) ActionDelegate;
		bit down;
		bit focus;
		ActionDelegate onClickDelegate;
	public:
		this( Application app0, DWnd parent, wchar[] label, ActionDelegate actionDelegate ) {
			super( app0, parent );
			addPaintHandler( WM_PAINT, &onPaint );
			addMouseDownHandler( WM_LBUTTONDOWN, &onMouseDown );
			addMouseUpHandler(   WM_LBUTTONUP,   &onMouseUp ); // release capture
			addMouseDownHandler( WM_RBUTTONDOWN, &onMouseDown );
			addMouseUpHandler(   WM_RBUTTONUP,   &onMouseUp ); // release capture
			down = false;
			focus = false;
			onClickDelegate = actionDelegate;
			wchar[] buf = label.dup;
			buf ~= '\0';
			SetWindowTextW( hwnd, cast(wchar*)buf );
		}
		this( Application parent ) { this( parent, parent, "", &onClick ); }
		this( Application parent, wchar[] label ) { this( parent, parent, label, &onClick ); }
		this( Application app0, DWnd parent ) { this( app0, parent, "", &onClick ); }
		this( Application app0, wchar[] label, ActionDelegate actionDelegate ) { this( app0, app0, label, actionDelegate ); }
		this( Application app0, DWnd parent, wchar[] label ) { this( app0, parent, label, &onClick ); }

		void setNotify( ActionDelegate actionDelegate ) {
			onClickDelegate = actionDelegate;
		}
		void onClick( UserButton self ) { }

	/* interface HasPreferedSize { */
		void getPreferedSize( out int width, out int height ) {
			SIZE size;
			HDC dc;
			width = 100;
			height = 32;
			int len = GetWindowTextLengthW( hwnd );
			wchar[] buf = new wchar[len];
			len = GetWindowTextW( hwnd, cast(wchar*)buf, len );
			dc = GetDC( hwnd );
			GetTextExtentPoint32W( dc, cast(wchar*)buf, len, &size );
			ReleaseDC( hwnd, dc );
			width  = size.cx;
			height = size.cy;
		}
	/* } interface HasPreferedSize */

		LRESULT onPaint( HDC dc ) {
			RECT rect;
			getClientRect( rect );
			drawTricolBox( dc, rect,
								app.getLolightOnePen(),
								app.getMidlightOnePen(),
								app.getHilightOnePen(),
								null );
			deflateRect( rect, 1, 1 );
			if ( down ) {
				if ( focus ) {
					drawTricolBox( dc, rect,
									app.getLolightTwoPen(),
									app.getMidlightTwoPen(),
									app.getHilightTwoPen(),
									app.getMidlightTwoBrush() );
				} else {
					drawTricolBox( dc, rect,
									app.getLolightOnePen(),
									app.getMidlightOnePen(),
									app.getHilightOnePen(),
									app.getMidlightOneBrush() );
				}
			} else {
				if ( focus ) {
					drawTricolBox( dc, rect,
								app.getHilightTwoPen(),
								app.getMidlightTwoPen(),
								app.getLolightTwoPen(),
								app.getMidlightTwoBrush() );
				} else {
					drawTricolBox( dc, rect,
								app.getHilightOnePen(),
								app.getMidlightOnePen(),
								app.getLolightOnePen(),
								app.getMidlightOneBrush() );
				}
			}
			deflateRect( rect, 2, 2 );

			IntersectClipRect( dc, rect.left, rect.top, rect.right, rect.bottom );

			int w = (rect.right - rect.left);
			int h = rect.bottom-rect.top;
			int len = GetWindowTextLengthW( hwnd );
			wchar[] buf = new wchar[len];
			len = GetWindowTextW( hwnd, cast(wchar*)buf, len );
			SIZE size;

			HGDIOBJ oldFont;
			oldFont = SelectObject( dc, app.getDefaultFont() );

			GetTextExtentPoint32W( dc, cast(wchar*)buf, len, &size );
			while ( size.cx > w ) {
				if ( len <= 3 ) { break; }
				len--;
				buf[len-3..len] = cast(wchar[])"...";
				GetTextExtentPoint32W( dc, cast(wchar*)buf, len, &size );
			}
			int x, y;
			x = rect.left + ((w-size.cx)>>1);
			y = rect.top  + ((h-size.cy)>>1);
/*
			rect.left = x; rect.right  = x+size.cx;
			rect.top  = y; rect.bottom = y+size.cy;
			HBRUSH brush = CreateSolidBrush( 0xFF0000 );
			FillRect( dc, &rect, brush );
			DeleteObject( brush );
*/
			drawStringCF( dc, x, y, 0, buf, len );
			SelectObject( dc, oldFont );
			return cast(LRESULT)0;
		}

		LRESULT onMouseDown( int x, int y, int keys ) {
			down = true;
			invalidate();
			return cast(LRESULT)0;
		}

		LRESULT onMouseUp( int x, int y, int keys ) {
			down = false;
			if ( onClickDelegate ) {
				onClickDelegate( this );
			}
			invalidate();
			return cast(LRESULT)0;
		}
	}

	class UserFill : AppControl.AppWnd {
	public:
		this( Application app0, DWnd parent ) {
			super( app0, parent );
			addPaintHandler( WM_PAINT, &onPaint );
		}
		this( Application parent ) { this( parent, parent ); }

		LRESULT onPaint( HDC dc ) {
			RECT rect;
			getClientRect( rect );
			drawTricolBox( dc, rect,
								app.getLolightOnePen(),
								app.getMidlightOnePen(),
								app.getHilightOnePen(),
								null );
			deflateRect( rect, 1, 1 );
			drawTricolBox( dc, rect,
								app.getHilightOnePen(),
								app.getMidlightOnePen(),
								app.getLolightOnePen(),
								app.getMidlightOneBrush() );

			return cast(LRESULT)0;
		}
	}
}
